"use strict";

window.onload = init;

let gl;

// dane o programie cieniowania
let shaderProgram =
{
    "id" : null,
    "uModelMatrix" : null,
    "uViewMatrix" : null,
    "uProjectionMatrix" : null
};

// dane o czworokacie
let quad =
{
    "vao" : null,
    "positionVBO" : null,
    "texCoordVBO" : null,
    "noVertices" : 0
};

let projectionMatrix = glMatrix.mat4.create();
let viewMatrix = glMatrix.mat4.create();

/*
* funkcja inicjuje WebGL oraz ustawia poczatkowe wartosci pewnych ustawien WebGL
*/
function init()
{
    initGL();
    setupShaders();
    setupBuffers();

    render();
}

/*
* funkcja pobiera kontekst WebGL oraz inicjuje pewne ustawienia WebGL
*/
function initGL()
{
    let canvas = document.getElementById( "gl-canvas" );

    gl = canvas.getContext( 'webgl2' );
    if( !gl )
        alert( "WebGL2 jest niedostępny!" );

    gl.viewport( 0, 0, canvas.width, canvas.height );
    gl.clearColor( 0.0, 0.0, 0.0, 1.0 );

    glMatrix.mat4.perspective( projectionMatrix, glMatrix.glMatrix.toRadian( 45.0 ), canvas.width / canvas.height, 0.1, 100.0 );
    glMatrix.mat4.lookAt( viewMatrix, [0.0, 0.0, 1.0], [0.0, 0.0, 0.0], [0.0, 1.0, 0.0] );
}

/*
* funkcja wczytuje shadery oraz pobiera lokalizacje atrybutow i zmiennych jednorodnych
*/
function setupShaders()
{
    shaderProgram.id = createProgram( gl, "vertex-shader", "fragment-shader" );

    if( !shaderProgram.id )
        throw new Error( "Shader program creation failed!" );

    shaderProgram.uModelMatrix = gl.getUniformLocation( shaderProgram.id, "uModelMatrix" );
    shaderProgram.uViewMatrix = gl.getUniformLocation( shaderProgram.id, "uViewMatrix" );
    shaderProgram.uProjectionMatrix = gl.getUniformLocation( shaderProgram.id, "uProjectionMatrix" );
}

/*
* funkcja tworzy obiekty VBO z modelami do wyrenderowania
*/
function setupBuffers()
{
    let quadVerts =
    [
        -1.0, -1.0, -2.45, 1.0,
        1.0, -1.0, -2.45, 1.0,
        -1.0, 1.0, -2.45, 1.0,
        1.0, 1.0, -2.45, 1.0
    ];

    let texCoord =
    [
        0.0, 0.0,
        1.0, 0.0,
        0.0, 1.0,
        1.0, 1.0
    ];

    quad.vao = gl.createVertexArray();
    gl.bindVertexArray( quad.vao );

    quad.positionVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, quad.positionVBO );
    gl.enableVertexAttribArray( 0 );
    gl.vertexAttribPointer( 0, 4, gl.FLOAT, false, 0, 0 );
    gl.bufferData( gl.ARRAY_BUFFER, new Float32Array( quadVerts ), gl.STATIC_DRAW );

    quad.texCoordVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, quad.texCoordVBO );
    gl.enableVertexAttribArray( 1 );
    gl.vertexAttribPointer( 1, 2, gl.FLOAT, false, 0, 0 );
    gl.bufferData( gl.ARRAY_BUFFER, new Float32Array( texCoord ), gl.STATIC_DRAW );

    quad.noVertices = 4;
}

/*
* funkcja renderuje scene
*/
function render()
{
    gl.clear( gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT );

    gl.useProgram( shaderProgram.id );

    gl.uniformMatrix4fv( shaderProgram.uProjectionMatrix, false, projectionMatrix );
    gl.uniformMatrix4fv( shaderProgram.uViewMatrix, false, viewMatrix );

    let quadModelMatrix = glMatrix.mat4.create();
    gl.uniformMatrix4fv( shaderProgram.uModelMatrix, false, quadModelMatrix );

    gl.bindVertexArray( quad.vao );
    gl.drawArrays( gl.TRIANGLE_STRIP, 0, quad.noVertices );
}
