"use strict";

window.onload = init;

let gl; // kontekst WebGL

// dane o programie cieniowania ze wzorem marmuru
let marblePattern =
{
    "id" : null,
    "uModelViewMatrix" : null,
    "uNormalMatrix" : null,
    "uProjectionMatrix" : null,
    "uLightPosition" : null,
    "uLightAmbient" : null,
    "uLightDiffuse" : null,
    "uLightSpecular" : null,
    "uMaterialAmbient" : null,
    "uMaterialDiffuse" : null,
    "uMaterialSpecular" : null,
    "uMaterialShininess" : null
};

// dane o programie cieniowania ze wzorem drzewa
let woodPattern =
{
    "id" : null,
    "uModelViewMatrix" : null,
    "uNormalMatrix" : null,
    "uProjectionMatrix" : null,
    "uLightPosition" : null,
    "uLightAmbient" : null,
    "uLightDiffuse" : null,
    "uLightSpecular" : null,
    "uMaterialAmbient" : null,
    "uMaterialDiffuse" : null,
    "uMaterialSpecular" : null,
    "uMaterialShininess" : null
};

// dane o zrodle swiatla
let light =
{
    "position" : glMatrix.vec4.fromValues( 0.0, 0.0, 5.0, 1.0 ), // w ukladzie swiata
    "ambient" : glMatrix.vec3.fromValues( 0.2, 0.2, 0.2 ),
    "diffuse" : glMatrix.vec3.fromValues( 1.0, 1.0, 1.0 ),
    "specular" : glMatrix.vec3.fromValues( 1.0, 1.0, 1.0 )
};

// dane o materiale
let material1 =
{
    "ambient" : glMatrix.vec3.fromValues( 0.250000, 0.207250, 0.207250 ),
    "diffuse" : glMatrix.vec3.fromValues( 1.000000, 0.829000, 0.829000 ),
    "specular" : glMatrix.vec3.fromValues( 0.296648, 0.296648, 0.296648 ),
    "shininess" : 11.264
};

let material2 =
{
    "ambient" : glMatrix.vec3.fromValues( 0.212500, 0.127500, 0.054000 ),
    "diffuse" : glMatrix.vec3.fromValues( 0.714000, 0.428400, 0.181440 ),
    "specular" : glMatrix.vec3.fromValues( 0.393548, 0.271906, 0.166721 ),
    "shininess" : 25.6
};

// dane o czajniczku
let teapot =
{
    "vao" : null,
    "positionVBO" : null, // id VBO zawierajacego wspolrzedne
    "normalVBO" : null, // id VBO zawierajacego normale
    "indicesVBO" : null, // id VBO zawierajacego indeksy wierzcholkow scianek
    "noIndices" : 0, // liczba wierzcholkow do wyrenderowania
    "translationVector" : [0.0, 0.0, 0.0],
    "scaleFactors" : [1.0, 1.0, 1.0],
    "render" : renderTeapot
};

let modelViewMatrix = glMatrix.mat4.create(); // macierz model-widok
let projectionMatrix = glMatrix.mat4.create(); // macierz rzutowania
let rotationAngles = [0.0, 0.0, 0.0]; // katy obrotu wzgledem poszczegolnych osi

let shader = marblePattern;
let material = material1;

/*
* funkcja inicjuje WebGL oraz ustawia poczatkowe wartosci pewnych ustawien WebGL
*/
function init()
{
    initGL();
    setupShaders();
    setupBuffers();

    initUI();

    window.requestAnimationFrame( render );
}

/*
* funkcja pobiera kontekst WebGL oraz inicjuje pewne ustawienia WebGL
*/
function initGL()
{
    let canvas = document.getElementById( "gl-canvas" );

    gl = canvas.getContext( 'webgl2' );
    if( !gl )
        alert( "WebGL2 jest niedostępny!" );

    gl.viewport( 0, 0, canvas.width, canvas.height );
    gl.clearColor( 0.7, 0.7, 0.7, 1.0 );
    gl.enable( gl.DEPTH_TEST );

    glMatrix.mat4.perspective( projectionMatrix, glMatrix.glMatrix.toRadian( 40.0 ), canvas.width / canvas.height, 0.1, 10.0 );
}

/*
* funkcja wczytuje shadery oraz pobiera lokalizacje atrybutow i zmiennych jednorodnych
*/
function setupShaders()
{
    marblePattern.id = createProgram( gl, "vertex-shader", "marble_fragment-shader" );

    if( !marblePattern.id )
        throw new Error( "Shader program creation failed!" );

    marblePattern.uModelViewMatrix = gl.getUniformLocation( marblePattern.id, "uModelViewMatrix" );
    marblePattern.uNormalMatrix = gl.getUniformLocation( marblePattern.id, "uNormalMatrix" );
    marblePattern.uProjectionMatrix = gl.getUniformLocation( marblePattern.id, "uProjectionMatrix" );

    marblePattern.uLightPosition = gl.getUniformLocation( marblePattern.id, "uLight.position" );
    marblePattern.uLightAmbient = gl.getUniformLocation( marblePattern.id, "uLight.ambient" );
    marblePattern.uLightDiffuse = gl.getUniformLocation( marblePattern.id, "uLight.diffuse" );
    marblePattern.uLightSpecular = gl.getUniformLocation( marblePattern.id, "uLight.specular" );

    marblePattern.uMaterialAmbient = gl.getUniformLocation( marblePattern.id, "uMaterial.ambient" );
    marblePattern.uMaterialDiffuse = gl.getUniformLocation( marblePattern.id, "uMaterial.diffuse" );
    marblePattern.uMaterialSpecular = gl.getUniformLocation( marblePattern.id, "uMaterial.specular" );
    marblePattern.uMaterialShininess = gl.getUniformLocation( marblePattern.id, "uMaterial.shininess" );

    //-------------------------------------

    woodPattern.id = createProgram( gl, "vertex-shader", "wood_fragment-shader" );

    if( !woodPattern.id )
        throw new Error( "Shader program creation failed!" );

    woodPattern.uModelViewMatrix = gl.getUniformLocation( woodPattern.id, "uModelViewMatrix" );
    woodPattern.uNormalMatrix = gl.getUniformLocation( woodPattern.id, "uNormalMatrix" );
    woodPattern.uProjectionMatrix = gl.getUniformLocation( woodPattern.id, "uProjectionMatrix" );

    woodPattern.uLightPosition = gl.getUniformLocation( woodPattern.id, "uLight.position" );
    woodPattern.uLightAmbient = gl.getUniformLocation( woodPattern.id, "uLight.ambient" );
    woodPattern.uLightDiffuse = gl.getUniformLocation( woodPattern.id, "uLight.diffuse" );
    woodPattern.uLightSpecular = gl.getUniformLocation( woodPattern.id, "uLight.specular" );

    woodPattern.uMaterialAmbient = gl.getUniformLocation( woodPattern.id, "uMaterial.ambient" );
    woodPattern.uMaterialDiffuse = gl.getUniformLocation( woodPattern.id, "uMaterial.diffuse" );
    woodPattern.uMaterialSpecular = gl.getUniformLocation( woodPattern.id, "uMaterial.specular" );
    woodPattern.uMaterialShininess = gl.getUniformLocation( woodPattern.id, "uMaterial.shininess" );
}

/*
* funkcja tworzy obiekty VBO z czajniczkiem
*/
function setupBuffers()
{
    teapot.vao = gl.createVertexArray();
    gl.bindVertexArray( teapot.vao );

    teapot.positionVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, teapot.positionVBO );
    gl.enableVertexAttribArray( 0 );
    gl.vertexAttribPointer( 0, 3, gl.FLOAT, false, 0, 0 );
    gl.bufferData( gl.ARRAY_BUFFER, teapotPositions, gl.STATIC_DRAW );

    teapot.normalVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, teapot.normalVBO );
    gl.enableVertexAttribArray( 1 );
    gl.vertexAttribPointer( 1, 3, gl.FLOAT, false, 0, 0 );
    gl.bufferData( gl.ARRAY_BUFFER, teapotNormals, gl.STATIC_DRAW );

    teapot.indicesVBO = gl.createBuffer();
    gl.bindBuffer( gl.ELEMENT_ARRAY_BUFFER, teapot.indicesVBO );
    gl.bufferData( gl.ELEMENT_ARRAY_BUFFER, teapotIndices, gl.STATIC_DRAW );

    teapot.noIndices = teapotIndices.length;

    calculateTranslationAndScale();
}

/*
* funkcja wyznacza wektor translacji i wspolczynnik skali tak, aby czajniczek
* znajdowal sie w poczatku ukladu wspolrzednych i miescil sie w szescianie o boku 2
*/
function calculateTranslationAndScale()
{
    let bbMin = [teapotPositions[0], teapotPositions[1], teapotPositions[2]];
    let bbMax = [teapotPositions[0], teapotPositions[1], teapotPositions[2]];
    let centroid = [0.0, 0.0, 0.0];
    let n = 0;

    for( let i = 0; i < teapotPositions.length; i += 3 )
    {
        centroid[0] += teapotPositions[i];
        centroid[1] += teapotPositions[i + 1];
        centroid[2] += teapotPositions[i + 2];
        ++n;

        for( let j = 0; j < 3; ++j )
        {
            if( teapotPositions[i + j] < bbMin[j] )
                bbMin[j] = teapotPositions[i + j];

            if( bbMax[j] < teapotPositions[i + j] )
                bbMax[j] = teapotPositions[i + j];
        }
    }

    centroid = centroid.map( v => v / n );
    teapot.translationVector = centroid.map( v => -v );

    let maxExtent = Math.max.apply( Math, [Math.abs(bbMax[0] - bbMin[0]), Math.abs(bbMax[1] - bbMin[1]), Math.abs(bbMax[2] - bbMin[2])] );
    let scale = 2.0 / maxExtent;
    teapot.scaleFactors = [scale, scale, scale];
}

/*
* funkcja oblicza iloczyn macierzy przez wektor, tzn. M v
* M - macierz
* v - wektor
*/
function multiplyMatrixByVector(M, v)
{
    let x = M[0] * v[0] + M[4] * v[1] + M[8] * v[2] + M[12] * v[3];
    let y = M[1] * v[0] + M[5] * v[1] + M[9] * v[2] + M[13] * v[3];
    let z = M[2] * v[0] + M[6] * v[1] + M[10] * v[2] + M[14] * v[3];
    let w = M[3] * v[0] + M[7] * v[1] + M[11] * v[2] + M[15] * v[3];

    return glMatrix.vec4.fromValues( x, y, z, w );
}

/*
* funkcja renderujaca scene
*/
function render()
{
    gl.clear( gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT );

    gl.useProgram( shader.id );

    glMatrix.mat4.lookAt( modelViewMatrix, [0, 0, 3], [0, 0, 0], [0, 1, 0] );

    let lightPositionEye = multiplyMatrixByVector( modelViewMatrix, light.position ); // pozycja swiatla w ukladzie oka

    gl.uniform4fv( shader.uLightPosition, lightPositionEye );
    gl.uniform3fv( shader.uLightAmbient, light.ambient );
    gl.uniform3fv( shader.uLightDiffuse, light.diffuse );
    gl.uniform3fv( shader.uLightSpecular, light.specular );

    gl.uniform3fv( shader.uMaterialAmbient, material.ambient );
    gl.uniform3fv( shader.uMaterialDiffuse, material.diffuse );
    gl.uniform3fv( shader.uMaterialSpecular, material.specular );
    gl.uniform1f( shader.uMaterialShininess, material.shininess );

    gl.uniformMatrix4fv( shader.uProjectionMatrix, false, projectionMatrix );

    glMatrix.mat4.scale( modelViewMatrix, modelViewMatrix, teapot.scaleFactors );
    glMatrix.mat4.rotateZ( modelViewMatrix, modelViewMatrix, rotationAngles[2] * Math.PI / 180.0 );
    glMatrix.mat4.rotateY( modelViewMatrix, modelViewMatrix, rotationAngles[1] * Math.PI / 180.0 );
    glMatrix.mat4.rotateX( modelViewMatrix, modelViewMatrix, rotationAngles[0] * Math.PI / 180.0 );
    glMatrix.mat4.translate( modelViewMatrix, modelViewMatrix, teapot.translationVector );

    let normalMatrix = glMatrix.mat3.create();
    glMatrix.mat3.normalFromMat4( normalMatrix, modelViewMatrix );

    gl.uniformMatrix4fv( shader.uModelViewMatrix, false, modelViewMatrix );
    gl.uniformMatrix3fv( shader.uNormalMatrix, false, normalMatrix );

    teapot.render();

    window.requestAnimationFrame( render );
}

/*
* funkcja rysujaca czajniczek
*/
function renderTeapot()
{
    gl.bindVertexArray( this.vao );
    gl.drawElements( gl.TRIANGLES, this.noIndices, gl.UNSIGNED_SHORT, 0 );
}